
@file:Suppress("unused")

package ru.dbotthepony.kommons.math

import kotlin.math.pow

/**
 * Bézier curve of two components, denoted by [t].
 *
 * Mathematically it is equal to linear interpolation.
 */
fun bezierCurve(t: Float, a0: Float, a1: Float): Float {
	return linearInterpolation(t, a0, a1)
}

/**
 * Bézier curve of two components, denoted by [t].
 *
 * Mathematically it is equal to linear interpolation.
 */
fun bezierCurve(t: Double, a0: Double, a1: Double): Double {
	return linearInterpolation(t, a0, a1)
}

/**
 * Bézier curve of three components, denoted by [t].
 *
 * Mathematically, it is computed using binomial of third order.
 *
 * If numerical stability is a priority, use [bezierCurveStrict].
 */
fun bezierCurve(t: Float, a0: Float, a1: Float, a2: Float): Float {
	if (t <= 0f)
		return a0
	else if (t >= 1f)
		return a2

	val inv = 1f - t

	return inv * inv * a0 + 2f * t * inv * a1 + t * t * a2
}

/**
 * Bézier curve of three components, denoted by [t].
 *
 * Mathematically, it is computed using binomial of third order.
 *
 * If numerical stability is a priority, use [bezierCurveStrict].
 */
fun bezierCurve(t: Double, a0: Double, a1: Double, a2: Double): Double {
	if (t <= 0.0)
		return a0
	else if (t >= 1.0)
		return a2

	val inv = 1.0 - t

	return inv * inv * a0 + 2.0 * t * inv * a1 + t * t * a2
}

/**
 * Bézier curve of four components, denoted by [t].
 *
 * Mathematically, it is computed using binomial of fourth order.
 *
 * If numerical stability is a priority, use [bezierCurveStrict].
 */
fun bezierCurve(t: Float, a0: Float, a1: Float, a2: Float, a3: Float): Float {
	if (t <= 0f)
		return a0
	else if (t >= 1f)
		return a3

	val inv = 1f - t

	return inv * inv * inv * a0 + 3f * t * inv * inv * a1 + 3f * t * t * inv * a2 + t * t * t * a3
}

/**
 * Bézier curve of four components, denoted by [t].
 *
 * Mathematically, it is computed using binomial of fourth order.
 *
 * If numerical stability is a priority, use [bezierCurveStrict].
 */
fun bezierCurve(t: Double, a0: Double, a1: Double, a2: Double, a3: Double): Double {
	if (t <= 0.0)
		return a0
	else if (t >= 1.0)
		return a3

	val inv = 1.0 - t

	return inv * inv * inv * a0 + 3.0 * t * inv * inv * a1 + 3.0 * t * t * inv * a2 + t * t * t * a3
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm recursively compute curves of lower order, until order of 2 is reached,
 * in which case [linearInterpolation] is used.
 */
fun bezierCurveStrict(t: Float, vararg values: Float): Float {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			@Suppress("name_shadowing")
			val values = values.clone()

			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm recursively compute curves of lower order, until order of 2 is reached,
 * in which case [linearInterpolation] is used.
 *
 * This function does not copy [values] and modify it directly.
 */
fun bezierCurveStrictDirect(t: Float, values: FloatArray): Float {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm recursively compute curves of lower order, until order of 2 is reached,
 * in which case [linearInterpolation] is used.
 */
fun bezierCurveStrict(t: Double, vararg values: Double): Double {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			@Suppress("name_shadowing")
			val values = values.clone()

			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm recursively compute curves of lower order, until order of 2 is reached,
 * in which case [linearInterpolation] is used.
 *
 * This function does not copy [values] and modify it directly.
 */
fun bezierCurveStrictDirect(t: Double, values: DoubleArray): Double {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		else -> {
			// construct prime - 1
			for (length in values.size - 2 downTo 0) {
				// search prime
				for (j in 0 .. length) {
					val point1 = values[j]
					val point2 = values[j + 1]

					values[j] = linearInterpolation(t, point1, point2)
				}
			}

			return values[0]
		}
	}
}

private val polynomialCacheF = arrayOfNulls<FloatArray>(60)
private val polynomialCacheD = arrayOfNulls<DoubleArray>(160)

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm compute pascal triangle' row using [pascalTriangleRowFloat], and then compute
 * Bézier curve using computed coefficients.
 *
 * Generally, this method is faster than [bezierCurveStrict], but lack numeric stability the latter offers.
 *
 * This function use cache for reasonable order of pascal triangle.
 */
fun bezierCurveBinomial(t: Float, vararg values: Float): Float {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		2 -> return bezierCurve(t, values[0], values[1])
		3 -> return bezierCurve(t, values[0], values[1], values[2])
		4 -> return bezierCurve(t, values[0], values[1], values[2], values[3])
	}

	val poly: FloatArray
	val trigRow = values.size - 1

	if (values.size >= 60) {
		poly = pascalTriangleRowFloat(trigRow)
	} else {
		if (polynomialCacheF[trigRow] == null) {
			polynomialCacheF[trigRow] = pascalTriangleRowFloat(trigRow)
		}

		poly = polynomialCacheF[trigRow]!!
	}

	var sum = 0f
	val inv = 1f - t

	for (i in values.indices) {
		sum += inv.pow(trigRow - i) * t.pow(i) * values[i] * poly[i]
	}

	return sum
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm compute pascal triangle' row using [pascalTriangleRowFloat], and then compute
 * Bézier curve using computed coefficients.
 *
 * Generally, this method is faster than [bezierCurveStrict], but lack numeric stability the latter offers.
 *
 * This function use cache for reasonable order of pascal triangle.
 *
 * This function does not modify [values], but also avoids copying.
 */
fun bezierCurveBinomialDirect(t: Float, values: FloatArray): Float {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		2 -> return bezierCurve(t, values[0], values[1])
		3 -> return bezierCurve(t, values[0], values[1], values[2])
		4 -> return bezierCurve(t, values[0], values[1], values[2], values[3])
	}

	val poly: FloatArray
	val trigRow = values.size - 1

	if (values.size >= 60) {
		poly = pascalTriangleRowFloat(trigRow)
	} else {
		if (polynomialCacheF[trigRow] == null) {
			polynomialCacheF[trigRow] = pascalTriangleRowFloat(trigRow)
		}

		poly = polynomialCacheF[trigRow]!!
	}

	var sum = 0f
	val inv = 1f - t

	for (i in values.indices) {
		sum += inv.pow(trigRow - i) * t.pow(i) * values[i] * poly[i]
	}

	return sum
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm compute pascal triangle' row using [pascalTriangleRowDouble], and then compute
 * Bézier curve using computed coefficients.
 *
 * Generally, this method is faster than [bezierCurveStrict], but lack numeric stability the latter offers.
 *
 * This function use cache for reasonable order of pascal triangle.
 */
fun bezierCurveBinomial(t: Double, vararg values: Double): Double {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		2 -> return bezierCurve(t, values[0], values[1])
		3 -> return bezierCurve(t, values[0], values[1], values[2])
		4 -> return bezierCurve(t, values[0], values[1], values[2], values[3])
	}

	val poly: DoubleArray
	val trigRow = values.size - 1

	if (values.size >= 160) {
		poly = pascalTriangleRowDouble(trigRow)
	} else {
		if (polynomialCacheD[trigRow] == null) {
			polynomialCacheD[trigRow] = pascalTriangleRowDouble(trigRow)
		}

		poly = polynomialCacheD[trigRow]!!
	}

	var sum = 0.0
	val inv = 1.0 - t

	for (i in values.indices) {
		sum += inv.pow(trigRow - i) * t.pow(i) * values[i] * poly[i]
	}

	return sum
}

/**
 * Bézier curve of any order, denoted by [t].
 *
 * This algorithm compute pascal triangle' row using [pascalTriangleRowDouble], and then compute
 * Bézier curve using computed coefficients.
 *
 * Generally, this method is faster than [bezierCurveStrict], but lack numeric stability the latter offers.
 *
 * This function use cache for reasonable order of pascal triangle.
 *
 * This function does not modify [values], but also avoids copying.
 */
fun bezierCurveBinomialDirect(t: Double, values: DoubleArray): Double {
	when (values.size) {
		0, 1 -> throw IllegalArgumentException("Provided array has only ${values.size} entries in it")
		2 -> return bezierCurve(t, values[0], values[1])
		3 -> return bezierCurve(t, values[0], values[1], values[2])
		4 -> return bezierCurve(t, values[0], values[1], values[2], values[3])
	}

	val poly: DoubleArray
	val trigRow = values.size - 1

	if (values.size >= 160) {
		poly = pascalTriangleRowDouble(trigRow)
	} else {
		if (polynomialCacheD[trigRow] == null) {
			polynomialCacheD[trigRow] = pascalTriangleRowDouble(trigRow)
		}

		poly = polynomialCacheD[trigRow]!!
	}

	var sum = 0.0
	val inv = 1.0 - t

	for (i in values.indices) {
		sum += inv.pow(trigRow - i) * t.pow(i) * values[i] * poly[i]
	}

	return sum
}
